module GenFold where
import ParseLib
import HaskellType
import List (intersperse)

genFold :: HaskellType -> String
genFold h = foldName ++ " :: " ++ dataType ++ " -> " ++ show dataConstr ++ " -> r" ++ "\n" ++
            foldName ++ " fs@" ++ foldArgs constrs ++ " = f fs where \n" ++ func
  where
   foldName = "fold" ++ dataName
   foldArgs ls = sCommaList $ variableList 'f' ls
   (dataName, constrs, dataConstr) = (\(HData dataConstr@(HConstructor name _) constrs) -> (name, constrs, dataConstr)) h
   dataType = foldType dataConstr constrs
   func = indent $ foldDef dataConstr constrs

indent = unlines . map ((++) "  " ) 

foldType :: HaskellType -> [HaskellType] -> String
foldType dataConstr xs = sParens $ concat $ intersperse "\n\t, " $ map typ xs
 where
  typ (HConstructor _ xs) = finishConstrTypes $ map showArg xs

  (dname, constrArgs) = (\(HConstructor name xs) -> (name, xs)) dataConstr

  finishConstrTypes [] = "r"
  finishConstrTypes xs = (++ "->r") $ concat $ intersperse "->" xs

  showArg :: HaskellType -> String
  showArg h@(HVariable hname) = show h
  showArg h@(HTypeConstructor hname) | hname == dname && constrArgs == [] = "r"
                                     | otherwise = show h
  showArg h@(HList list) = sBrackets $ showArg list
  showArg h@(HFunction arg1 arg2) = sParens $ showArg arg1 ++ "->" ++ show arg2
  showArg h@(HTupel tupel) = sCommaList $ map showArg tupel
  showArg h@(HTypeApplication arg1 arg2) | dataConstr == (toConstr h) = "r"
                                         | otherwise = show h

toConstr xs = HConstructor ((\(HTypeConstructor n) -> n) $ h) t where
  (h:t) = tapFoldl xs
  tapFoldl :: HaskellType -> [HaskellType]
  tapFoldl (HTypeApplication arg1 arg2) = tapFoldl arg1 ++ [arg2]
  tapFoldl x = [x]

variableList :: Char -> [HaskellType] -> [String]
variableList c ls = take (length ls ) $ map ((:) c . show) $ iterate (+1) 1

foldDef :: HaskellType -> [HaskellType] -> [String]
foldDef dataConstr xs = map (\(a,b)-> a b) $ functionList
 where
  functionList =  zip (map makeFunction xs) (variableList 'f' xs)

  makeFunction h@(HConstructor name xs) fName = "f fs " ++ fPattern ++ " = " ++ fBody fName
   where
    fPattern = sParens $ fDefinition' type2pattern name
    (dname, constrArgs) = case dataConstr of 
                           (HConstructor name xs) -> (name, xs)

    containsRecursion = or . map isRecursive 
    isRecursive h@(HTypeApplication _ _) = dataConstr == (toConstr h)
    isRecursive h@(HTypeConstructor x1) = x1 == dname && constrArgs == []
    isRecursive _ = False

    fBody = fDefinition' ttfb
    fDefinition' func name = unwords $ ((:) name . fst) $ foldl func ([], 1) xs

    simpleVar :: IntermediateState -> IntermediateState
    simpleVar (s,i) = (s ++ ['x': show i], i+1)

    type2pattern :: IntermediateState -> HaskellType -> IntermediateState
    type2pattern (s,i) t@(HTupel l) = handleTupel (s, i) t type2pattern
    type2pattern x _ = simpleVar x


    -- converts a type to the respective call in the function body
    -- ttfb = type to function body
    ttfb :: IntermediateState -> HaskellType -> IntermediateState
    ttfb s t@(HTupel l) = handleTupel s t ttfb
    ttfb s t@(HTypeApplication arg1 arg) | isRecursive t = simpleRecursion s
                                         | otherwise = simpleVar s
    ttfb s t@(HList list)                | isRecursive list = mapRecursion s
                                         | otherwise = simpleVar s
    ttfb s t@(HTypeConstructor x1)       | isRecursive t = simpleRecursion s
                                         | otherwise = simpleVar s
    ttfb s _ = simpleVar s

    simpleRecursion (s, i) = (s ++ [sParens $ "f fs x" ++ show i], i+1)
    mapRecursion (s, i) = (s ++ [sParens $ "map (f fs) x" ++ show i], i+1)

    handleTupel (s, i) t@(HTupel l) fn | containsRecursion l =  let (s', i') = foldl fn ([],i) l in 
                                                                (s ++ [(sCommaList s')], i')
                                       | otherwise = simpleVar (s,i)


type IntermediateState = ([String], Int)
