module Lexer where
import HaskellType
import ParseLib
import Char

-- datatype voor Tokens
data Token = LParen
           | RParen
           | LBrack
           | RBrack
           | Arrow
           | Comma
           | Equals
           | Bar
           | Data
           | Const String
           | Varia String
           deriving (Show, Eq)

-- onderdeel a
lowerIdent :: Parser Char String
lowerIdent = ident isLower

upperIdent :: Parser Char String
upperIdent = ident isUpper

whiteSpace :: Parser Char String
whiteSpace = greedy $ satisfy isSpace

ident :: (Char -> Bool) -> Parser Char String
ident f = list <$> satisfy f <*> idRest
 where
  idRest :: Parser Char String
  idRest = greedy $ satisfy isAlphaNum <|> symbol '\''

-- onderdeel b
terminals =
    [ (LParen, "("   )
    , (RParen, ")"   )
    , (LBrack, "["   )
    , (RBrack, "]"   )
    , (Arrow,  "->"  )
    , (Comma,  ","   )
    , (Equals, "="   )
    , (Bar,    "|"   )
    , (Data,   "data")
    ]

parseToken :: Parser Char Token
parseToken =   choice ( map tupel2Parser terminals )
           <|> Varia <$> lowerIdent
           <|> Const <$> upperIdent
  where
   tupel2Parser :: (Token, String) -> Parser Char Token
   tupel2Parser (t, s) = const t <$> token s
 
-- onderdeel c
lexicalScanner :: Parser Char [Token]
lexicalScanner = pack whiteSpace (listOf parseToken whiteSpace) whiteSpace