module Parser where
import HaskellType
import ParseLib
import Lexer

-- onderdeel e
parseType :: Parser Token HaskellType
parseType =   parseBType
          <|> HFunction <$> parseBType <*> ( pi22 <$> symbol Arrow <*> parseType )

parseBType :: Parser Token HaskellType
parseBType = foldl1 HTypeApplication <$> many1 parseAType

parseAType :: Parser Token HaskellType
parseAType =   parseVariable
           <|> HTypeConstructor <$> parseConstName
           <|> makeTupel <$> pack (symbol LParen) (listOf parseType (symbol Comma) ) (symbol RParen)
           <|> HList <$> pack (symbol LBrack) parseType (symbol RBrack)
  where
   makeTupel :: [HaskellType] -> HaskellType
   makeTupel [x] = x
   makeTupel xs  = HTupel xs

-- Onderdeel f

parseVariable :: Parser Token HaskellType
parseVariable = (\(Varia x) -> HVariable x) <$> satisfy isVaria
  where
   isVaria :: Token -> Bool
   isVaria (Varia x) = True
   isVaria _ = False

parseConstName :: Parser Token String
parseConstName = (\(Const x) -> x) <$> satisfy isConst
  where
   isConst :: Token -> Bool
   isConst (Const x) = True
   isConst _ = False

parseConstructor :: Parser Token HaskellType
parseConstructor = HConstructor <$> parseConstName <*> many parseAType

parseConstructors :: Parser Token [HaskellType]
parseConstructors = listOf parseConstructor (symbol Bar)

parseSimpleType :: Parser Token HaskellType
parseSimpleType = HConstructor <$> parseConstName <*> many parseVariable

parseData :: Parser Token HaskellType
parseData = makeData <$> symbol Data <*> parseSimpleType <*> symbol Equals <*> parseConstructors
  where 
   makeData _ t _ c = HData t c

parseInvoer :: Parser Token [HaskellType]
parseInvoer = many parseData

